// This code is licensed under GPL v3
// see license.txt for details

#include <LiquidCrystal.h>

#define DATA_IDLE 0
#define DATA_RECORD_START 1
#define DATA_RECORD_NAME 2
#define DATA_RECORD_VALUE 3
#define DATA_RECORD_HEX 4
#define DATA_CHECKSUM 5
#define EXIT_INCOMPLETE 0
#define EXIT_OK 1
#define EXIT_WRONG_CHECKSUM 2
#define EXIT_HEX 3
#define NAME_LENGTH 10
#define VALUE_LENGTH 34
#define NAME_VALUE_BUFFER 18
#define LOAD_PIN 8

const float UNDERVOLTAGE_THRESHOLD = 23.5; /* Volts */
const unsigned long FLOATTIME_LOWER_THRESHOLD = 3600; /* Seconds per week, so 3600 is one hour */
const unsigned long FLOATTIME_UPPER_THRESHOLD = 30 * 3600; /* 30 hours */
const unsigned long CHECK_PERIOD = 1000; /* one second */
const unsigned char HAPPY_THRESHOLD = 28; /* Hours per week, battery ist happy */
const unsigned char NEUTRAL_THRESHOLD = 14; /* Not that happy,but still o.k. 
                                               Should be greater than the maximum possible daily floating time in your region,
                                               otherwise the warning time (unhappy face) might be (worst case) only one day */

// Victron buffer & stuff
char names[NAME_VALUE_BUFFER][NAME_LENGTH];
char values[NAME_VALUE_BUFFER][VALUE_LENGTH];
bool firstByte = true; /* New string from the charger */

// Various strings for the display
char voltageString[5] = "";
char currentString[5] = "";
char powerString[5] = "";
char yieldTodayString[5] = "";
char floatHoursString[5] = "";

bool load = true; /* Load output state for 75/10, 75/15, 100/15, 100/20, currently unused */
char chargerStateByte = 'U'; /* means "unknown" */
float lastYieldToday = 0; /* Used for daily turnover, detect a new day */
unsigned long floatSecondsArray[7] = { 3600, 3600, 3600, 3600, 3600, 3600, 3600 }; /* Init, one hour per day */
bool isFloating = false;
unsigned long CheckPeriod; /* For periodic tasks */
unsigned long floatSecondsThisWeek = 0;
bool voltageOk = true; /* Is the voltage o.k.? */
bool floatOk = true; /* The same, floating time per week */
float voltage = UNDERVOLTAGE_THRESHOLD + 1.0; /* Initialization, avoid immediate load shutdown */

static unsigned char happy[8] = { /* Icons for visualization */
  0b00000,
  0b00000,
  0b01010,
  0b00000,
  0b00000,
  0b10001,
  0b01110,
  0b00000
};

static unsigned char unhappy[8] = {
  0b00000,
  0b00000,
  0b01010,
  0b00000,
  0b00000,
  0b01110,
  0b10001,
  0b00000
};

static unsigned char neutral[8] = {
  0b00000,
  0b00000,
  0b01010,
  0b00000,
  0b00000,
  0b11111,
  0b00000,
  0b00000
};


LiquidCrystal lcd(12, 11, 5, 4, 3, 2);

void setup() {
  Serial.begin(19200);
  Serial.setTimeout(10);
  lcd.begin(16,2);
  lcd.createChar(0, happy); // Create characters for visualization
  lcd.createChar(1, neutral);
  lcd.createChar(2, unhappy);
  lcd.clear();
  
  for (int i = 0; i < 7; i++) {
    floatSecondsThisWeek += floatSecondsArray[i]; // Calculate initial weekly floating time from the defaults
  }
  pinMode(LOAD_PIN, OUTPUT); /* Initialize load pin */
  CheckPeriod = millis() + CHECK_PERIOD; 
}

char getVictronData(char inbyte) {
  const char checksumName[] = "CHECKSUM";
  static char dataState;
  static char checksum;
  static char *textPointer;
  static char nameValuePair;
  char exitState = EXIT_INCOMPLETE;
  static char valueName[NAME_LENGTH];
  static char value[VALUE_LENGTH];

  // Initialize buffer et al
  if (firstByte) {
    for (int bufferNumber = 0; bufferNumber < NAME_VALUE_BUFFER; bufferNumber++) {
      strcpy(names[bufferNumber], "");
      strcpy(values[bufferNumber], "");
    }
    dataState = DATA_IDLE;
    checksum = 0;
    nameValuePair = 0;
    firstByte = false;
  }


  if ((inbyte == ':') && (dataState != DATA_CHECKSUM)) {
    dataState = DATA_RECORD_HEX;
  }
  if (dataState != DATA_RECORD_HEX) {
    checksum += inbyte;
  }
  inbyte = toupper(inbyte);
  
  switch(dataState) {
  case DATA_IDLE:
    /* wait for \n of the start of a record */
    switch(inbyte) {
    case '\n':
      dataState = DATA_RECORD_START;
      break;
    case '\r': /* Probably some MS-stuff, discard */
    default:
      break;
    }
    break;
  case DATA_RECORD_START:
    // Set pointer and store first byte, change dataState
    textPointer = valueName;
    *textPointer++ = inbyte;
    dataState = DATA_RECORD_NAME;
    break;
  case DATA_RECORD_NAME:
    // The record name is being received, terminated by a \t
    switch(inbyte) {
    case '\t':      
      if ( textPointer < (valueName + sizeof(valueName)) ) {
        *textPointer = 0; /* C-String, zero terminated */
        // We have reached the checksum
        if (strcmp(valueName, checksumName) == 0) {
          dataState = DATA_CHECKSUM;
          break;
        }
      }
      textPointer = value; /* Reset value pointer */
      dataState = DATA_RECORD_VALUE;
      break;
    default:
      // add byte to name, but do no overflow
      if ( textPointer < (valueName + sizeof(valueName)) )
        *textPointer++ = inbyte;
      break;
    }
    break;
    case DATA_RECORD_VALUE:

    // The record value is being received.  The \r indicates a new record.
    switch(inbyte) {
    case '\n':
      // forward record, only if it could be stored completely
      if ( textPointer < (value + sizeof(value)) ) {
        *textPointer = 0; // make zero ended
        // Store name and value into array
        strcpy(names[nameValuePair], valueName);
        strcpy(values[nameValuePair], value);
        if (nameValuePair < NAME_VALUE_BUFFER) {   /* Don't overflow */
          nameValuePair++;
        }
      }
      dataState = DATA_RECORD_START;
      break;
    case '\r': /* Skip */
      break;
    default:
      // add byte to value, but do no overflow
      if ( textPointer < (value + sizeof(value)) )
        *textPointer++ = inbyte;
      break;
    }
    break;
  case DATA_CHECKSUM:
  { 
    bool valid = checksum == 0; // All bytes of the message modulo 256 has to be zero
    if (!valid) {
      exitState = EXIT_WRONG_CHECKSUM;
    }
    else {
      exitState = EXIT_OK;
    }
    
    break;
  }
  case DATA_RECORD_HEX:
    if (inbyte == '\n') {
      exitState = EXIT_HEX;
    }
    break;
  }

  switch(exitState) {
  case EXIT_INCOMPLETE:
    return 0;
    break;
  case EXIT_OK:
    firstByte = true; 
    return 1;
    break;
  case EXIT_WRONG_CHECKSUM:
    firstByte = true;
    return 0;
    break;
  case EXIT_HEX:
    firstByte = true; 
    return 0;
    break;
  }
}


void loop() {
  char outputString1[17], outputString2[17];
  char newByte;
  
  if ((long)(millis() - CheckPeriod) >= 0) { /* What to do every second */
    if (isFloating) {
      floatSecondsArray[0]++;  // Update today's floating time
      floatSecondsThisWeek++;  // Update weekly floating time
      if (!floatOk && (floatSecondsThisWeek >= FLOATTIME_UPPER_THRESHOLD)) { /* Did we reach the floating time threashold? */
        floatOk = true;
      }
      if (!voltageOk) { /* Undervoltage state? If we reached float again, unset it */
        voltageOk = true;
      }
    }
    else {
      if ((voltage < UNDERVOLTAGE_THRESHOLD) && voltageOk) { /* Not floating? Check for undervoltage */
        voltageOk = false;
      }
    }
    digitalWrite(LOAD_PIN, (floatOk && voltageOk));
    CheckPeriod += CHECK_PERIOD;
  }
  
  if (Serial.available() > 0) {
    newByte = Serial.read();
    if (getVictronData(newByte) > 0) {
      for (int i=0; i < NAME_VALUE_BUFFER; i++) {   
        if (!strcmp(names[i], "V")) { // Voltage (mV)
          voltage = ((float)atol(values[i]) / (float)1000);
          dtostrf(voltage, 4, 1, voltageString);
          }
        else if (!strcmp(names[i], "I")) { // Current (mA)
          float current = ((float)atol(values[i]) / (float)1000);
          dtostrf(current, 4, 1, currentString);
        }
        else if (!strcmp(names[i], "PPV")) { // current PV-output (W)
          sprintf(powerString, "%4i", atoi(values[i]));
        }
        else if (!strcmp(names[i], "H20")) { // daily yield (10 Wh)
          float yieldToday = ((float)atol(values[i]) / (float)100);
          if (lastYieldToday > yieldToday) {  // Yield reset, so it's a new day
            floatSecondsThisWeek = 0; // Reset counter
            for (int i = 6; i > 0; i--) {
              floatSecondsArray[i] = floatSecondsArray[i-1]; // Move daily values
              floatSecondsThisWeek += floatSecondsArray[i]; // Calculate new weekly floating time
            }
            floatSecondsArray[0] = 0; // Reset today's counter
            if (floatSecondsThisWeek <= FLOATTIME_LOWER_THRESHOLD) { /* Did we reach the threshold for the weekly floating time= */ 
              floatOk = false;
            }
          }
          lastYieldToday = yieldToday;
          dtostrf(yieldToday, 4, 2, yieldTodayString);
        }
        else if (!strcmp(names[i], "LOAD")) { // Load (ON/OFF)
          if (!strcmp(values[i], "ON")) {
            load = true;
          }
          else if (!strcmp(values[i], "OFF")) {
            load = false;
          }
        }
        else if (!strcmp(names[i], "CS")) { // State, 0=off, 2=fault, 3=bulk, 4=absorption, 5=float
            
            unsigned char chargerState = atoi(values[i]);
          
            switch (chargerState) {
            case 0:
              chargerStateByte = 'O'; // Off
              isFloating = false;
              break;
            case 2: 
              chargerStateByte = 'E'; // Fault
              isFloating = false;
              break;
            case 3:
              chargerStateByte = 'B'; // Bulk
              isFloating = false;
              break;
            case 4:
              chargerStateByte = 'A'; // Absorption
              isFloating = false;
              break;
            case 5:
              chargerStateByte = 'F'; // Float
              isFloating = true;
              break;
            default:
              chargerStateByte = 'U'; // Unknown
              isFloating = false;
              break;       
          }
        }
      }    

      unsigned char floatHoursThisWeek = floatSecondsThisWeek / 3600;
      sprintf(outputString1, "% 4sV % 4sA % 3ih", voltageString, currentString, floatHoursThisWeek);
      if (voltageOk && floatOk) {
        sprintf(outputString2, "% 4sW % 4skWh   ", powerString, yieldTodayString);      
      }
      else if (floatOk) {
        sprintf(outputString2, "Undervoltage   ");
      }
      else if (voltageOk) {
        sprintf(outputString2, "Underfloat     ");
      }
      else {
        sprintf(outputString2, "Underf/Underv  ");
      }
      lcd.setCursor(0,0);
      lcd.print(outputString1);
      lcd.setCursor(0,1);
      lcd.print(outputString2);
      lcd.setCursor(14,1);
      if (floatHoursThisWeek >= HAPPY_THRESHOLD) {
        lcd.write((byte)0);
      }
      else if (floatHoursThisWeek >= NEUTRAL_THRESHOLD) {
          lcd.write((byte)1);
      }
      else {
        lcd.write((byte)2);
      }
      lcd.setCursor(15,1);
      lcd.write(chargerStateByte);
    }
  }
}
